# -*- coding: utf-8 -*-
"""
Created on Thu Apr 30 09:10:51 2020
@author: Luja von Köckritz

"""

# define imports, counters & settings
import pandas as pd
import numpy as np
import sys, os
#import matplotlib.pyplot as plt
#from past.builtins import execfile

os.chdir("C:\\Users\\Admin\\surfdrive\\Paper_5\\Luja\\furniture_model\\")
#os.chdir("C:\\Users\\luja\\Documents\\GSS\\Undergraduate_research_project\\furniture_model\\")
#datapath = 'C:\\Users\\luja\\Documents\\GSS\\Undergraduate research project\\furniture_model\\furniture_data'
from dynamic_stock_model import DynamicStockModel as DSM

#execfile('read_mym.py')         # module to read in mym (IMAGE output) files
idx = pd.IndexSlice             # needed for slicing multi-index dataframes (pandas)
scenario = "SSP2"
variant  = "BL"                # switch to "BL" for Baseline

#path = 'C:\\Users\\luja\\Documents\\GSS\\Undergraduate_research_project\\furniture_model\\'
path = 'C:\\Users\\Admin\\surfdrive\\Paper_5\\Luja\\furniture_model\\'
scenpath = scenario + "\\" +  scenario + "_" + variant + "\\"
datapath = path + 'furniture_data\\'
socialpath = path + 'social_data\\'
imagepath = path + scenpath

startyear = 1971
endyear = 2100
regions = 26
regions_list = ['1','2','3','4','5','6','7','8','9','10','11','12','13','14','15','16','17','18','19','20','21','22','23','24','25','26']
regions_names =  ['Canada','USA','Mexico','Central America', 'Brazil','Rest of South America','Northern Africa','Western Africa','Eastern Africa','South Africa','Western Europe','Central Europe','Turkey','Ukraine region','Central Asia','Russia region','Middle East','India','Korea region','China region','Southeastern Asia','Indonesia region','Japan','Oceania','Rest of South Asia','Rest of Southern Africa']
#regions_dict = dict(zip(regions_list, regions_names))
years = list(range(startyear,endyear+1))
quintiles = ['Q1','Q2','Q3','Q4','Q5']
#material_select = 'Steel'

#%% Define fucntions

# some csv files are defined only for a few years, this function ensures that a linear interpolation us used to derive data for every year between 1971 and 2100
def interpolate_data(data):
    data = data.append(pd.Series(data.loc[data.last_valid_index()], name=endyear))   
    data = data.append(pd.Series(data.loc[data.first_valid_index()], name=startyear))

    for i in range(startyear, endyear):
        if data.index.contains(i):
            pass
        else:
            data = data.append(pd.Series(name=i))
    
    data = data.sort_index(axis=0).interpolate(axis=0)
    return data

# In order to calculate inflow & outflow smoothly (without peaks for the initial years), we calculate a historic tail to the stock, by adding a 0 value for first year of operation (=1926), then interpolate values towards 1971
# expected input: columns: regions; index: years 
def stock_tail(stock):
    zero_value = [0 for i in range(0,regions)]
    stock_used = pd.DataFrame(stock).reindex_like(stock)
    stock_used.loc[first_year_furniture] = zero_value  # set all regions to 0 in the year of initial operation
    stock_new  = stock_used.reindex(list(range(first_year_furniture,endyear+1))).interpolate()
    return stock_new

# first define a Function in which the stock-driven DSM is applied to return (the moving average of the) inflow & outflow for all regions
def inflow_outflow(stock, lifetime):

    initial_year = stock.first_valid_index()
    outflow = pd.DataFrame(index=range(startyear,endyear+1), columns=stock.columns)
    inflow =  pd.DataFrame(index=range(startyear,endyear+1), columns=stock.columns)

    # define mean & standard deviation
    mean_list = [lifetime for i in range(0,len(stock))]   
    stdev_list = [mean_list[i] * stdev_mult for i in range(0,len(stock))]  

    for region in list(stock.columns):
        # define and run the DSM                                                                                            # list with the fixed (=mean) lifetime of grid elements, given for every timestep (1926-2100), needed for the DSM as it allows to change lifetime for different cohort (even though we keep it constant)
        DSMforward = DSM(t = np.arange(0,len(stock[region]),1), s=np.array(stock[region]), lt = {'Type': 'FoldedNormal', 'Mean': np.array(mean_list), 'StdDev': np.array(stdev_list)})  # definition of the DSM based on a folded normal distribution
        out_sc, out_oc, out_i = DSMforward.compute_stock_driven_model(NegativeInflowCorrect = True)                                                                 # run the DSM, to give 3 outputs: stock_by_cohort, outflow_by_cohort & inflow_per_year
    
        # sum the outflow by cohort to total outflow & apply a 5yr moving average to return only the 'smoothed' results
        out_o = out_oc.sum(axis=1)
        out_o = np.append(out_o,[out_o[-1],out_o[-1]])  # append last value twice, to get a more accurate moving average
        out_i = np.append(out_i,[out_i[-1],out_i[-1]])  # append last value twice, to get a more accurate moving average
    
        # apply moving average & return only 1971-2100 values
        outflow[region] = pd.Series(out_o[2:], index=list(range(initial_year,2101))).rolling(window=5).mean().loc[list(range(1971,2101))]    # Apply moving average                                                                                                      # sum the outflow by cohort to get the total outflow per year
        inflow[region] = pd.Series(out_i[2:], index=list(range(initial_year,2101))).rolling(window=5).mean().loc[list(range(1971,2101))]
       
    return inflow, outflow

#%% Read in files --------------------------------------------------------------------------

# IMAGE data
# Interpolate population in millions
raw_pop_data = pd.read_csv(imagepath + 'pop.csv', delimiter=';', index_col = 0)
new_pop_data = raw_pop_data.reindex(years).interpolate()
new_pop_data.columns = regions_names

# Import share of rural population per region oer year (IMAGE data)
rur_share = pd.read_csv(imagepath + 'rurpop.csv', delimiter=',', index_col = 0)
rur_share.columns = regions_names
rur_share = interpolate_data(rur_share)
rur_share = rur_share.drop(1970, axis=0)

# Calculate rural and urban population in millions
pop_rur = pd.DataFrame().reindex_like(rur_share)
pop_urb = pd.DataFrame().reindex_like(rur_share)
pop_rur = new_pop_data.multiply(rur_share, axis = 'index')
pop_urb = new_pop_data.multiply(1-rur_share, axis = 'index')
pop_rur = pop_rur.stack()
pop_urb = pop_urb.stack()

# Calculate floor space (m2) per capita for each quintile - split into rural and urban floorspace
raw_fli_data = pd.read_csv(imagepath + 'res_Floorspace.csv', delimiter=';', skiprows = 2, index_col = [0,1])
raw_fli_data.drop(['Total', 'Urban', 'Rural'], axis=1, inplace=True)
raw_fli_data.drop(raw_fli_data.loc[idx[:,27],:].index, inplace=True)
raw_fli_data = raw_fli_data.stack()
raw_fli_data = raw_fli_data.unstack(1)
raw_fli_data.columns = regions_names                # Might be better to do this with a dictionary
raw_fli_data=raw_fli_data.stack().unstack(1)
raw_fli_data = raw_fli_data.rename_axis(index=['t','Region'])

# read in additional (non-IMAGE) furniture data files
materials_raw = pd.read_csv(datapath + 'materials_bare.csv', delimiter=',', index_col = [0,1])       # kg per unit
f_amount_raw = pd.read_csv(datapath + 'furniture_amounts_bare.csv', delimiter=';', index_col = [0])  # unit per capita per m² for each quintile
lifetimes = pd.read_csv(datapath + 'lifetimes.csv', delimiter=';', index_col = [0])         # years
poverty_gap = pd.read_csv(socialpath + 'poverty_gap.csv', delimiter=',', index_col = [0])     # poverty gap by region in % of population living below US$ 1.90 / day 
gini_index = pd.read_csv(socialpath + 'gini_index.csv', delimiter=',', index_col = [0])       # GINI index (0-100) by region

# Reindexing poverty_gap and gini_index to the same region names format as the other datasets 
reindex_d = pd.DataFrame(columns=['Region data','Region model'])
reindex_d['Region model'] = regions_names
reindex_d['Region data'] = poverty_gap.columns.values.tolist()
reindex = reindex_d.set_index('Region data').to_dict()
poverty_gap.columns = poverty_gap.columns.to_series().map(reindex['Region model'])
gini_index.columns = gini_index.columns.to_series().map(reindex['Region model'])

# interpolate poverty gap and gini index, reindex to regions names used in the other files and drop the additional year 1970 
poverty_gap = interpolate_data(poverty_gap)
poverty_gap = poverty_gap.reindex(sorted(poverty_gap.columns), axis=1)
poverty_gap = poverty_gap.drop(1970, axis=0)
gini_index = interpolate_data(gini_index)
gini_index = gini_index.reindex(sorted(gini_index.columns), axis=1)
gini_index = gini_index.drop(1970, axis=0)
#%% Pre-calculations of the furniture sufficency & second hand fraction

# settings for the income based sufficiency calculations 
# Sufficiency is a multiplication factor to define how much less or more furniture you have compared to an average inventory of basic furniture items
# it is combined with the items per square meter, so a sufficiency < 1 means you don't own everything you might want to have, and a sufficiency >1 means you own more than you need on a per square meter basis
suf_max_q1 = 0.8   # maximum sufficiency of q1
suf_20_q1  = 0.5   # sufficiency of q1 at 20% poverty gap (= the entire first quinitile is below the poverty line)
suf_min_q5 = 1.05  # minimum sufficiency of q5
suf_max_q5 = 1.25  # maximum sufficiency of q5 
suf_gap    = 0.25  # gap between q2/q4 and q1/q5 (Q2 & Q4 are intermediate quintiles and their deviation from 1 is defined in relation to how low q1 goes and how high q5 goes), defined as the fraction of the gap with 1 remaining in the quintile 

#pre-calculate step-size 
suf_step_q1 = (suf_max_q1-suf_20_q1)/20
suf_step_q5 = (suf_max_q5-suf_min_q5)/(gini_index.max().max()-gini_index.min().min())

#define and calculate the final sufficiency dataframe (first Q1 & Q5, then Q2 & Q4)
index = pd.MultiIndex.from_product([poverty_gap.index, quintiles], names=['years', 'quintiles'])
sufficiency = pd.DataFrame(index=index, columns=poverty_gap.columns)

sufficiency.loc[idx[:,'Q1'],:] = suf_max_q1-(suf_step_q1 * poverty_gap).to_numpy() 
sufficiency.loc[idx[:,'Q3'],:] = 1 
sufficiency.loc[idx[:,'Q5'],:] = suf_min_q5 + (gini_index-gini_index.min().min()).to_numpy() * suf_step_q5
sufficiency.loc[idx[:,'Q2'],:] = 1 - (1 - sufficiency.loc[idx[:,'Q1'],:]).to_numpy() * suf_gap
sufficiency.loc[idx[:,'Q4'],:] = 1 + (sufficiency.loc[idx[:,'Q5'],:] - 1).to_numpy() * suf_gap

# settings for the income based calculations of the second hand ownership as a fraction (of the stock)
sec_min_q1   = 0.1
sec_20_q1    = 0.5
sec_step_q1  = (sec_20_q1-sec_min_q1)/20    # 20 because a quintile contains 20% of the people, step is the increase of second hand per %-point of the poverty rate
sec_decrease = 0.5                          # For each increase in quintile, how much less second hand furniture is used (with respect to a pre-defined q1)

#define and calculate the second hand fraction in the stock 
second_hand = pd.DataFrame().reindex_like(sufficiency)

second_hand.loc[idx[:,'Q1'],:] = sec_min_q1 + (poverty_gap.to_numpy() * sec_step_q1)
second_hand.loc[idx[:,'Q2'],:] = second_hand.loc[idx[:,'Q1'],:].to_numpy() * sec_decrease
second_hand.loc[idx[:,'Q3'],:] = second_hand.loc[idx[:,'Q2'],:].to_numpy() * sec_decrease
second_hand.loc[idx[:,'Q4'],:] = second_hand.loc[idx[:,'Q3'],:].to_numpy() * sec_decrease
second_hand.loc[idx[:,'Q5'],:] = second_hand.loc[idx[:,'Q4'],:].to_numpy() * sec_decrease

#%% Calculations
# Calculate material demand per floorspace per quintile
f_amount_raw = f_amount_raw.reindex(sorted(f_amount_raw.columns), axis=1)  # not sure if this is neesed
furniture_list=f_amount_raw.index.tolist()
#material_list=materials.index.get_level_values(1).unique().tolist()
#in_mat=materials.index
col_mat=materials_raw.columns
f_index = pd.MultiIndex.from_product([years, regions_names, quintiles], names=['t', 'Region','Quintile'])

# Calculate total weight of materials in furniture (kg per unit)
furniture_weight_raw = materials_raw.sum(axis=0, level=0)

#Relative differences in weight between quintiles
weight_factor = pd.Series([0.783479305, 0.9114296142, 1, 1.22834341, 1.795574561], index=quintiles) # Data collected from IKEA.nl, Meubis.nl, jysk.nl in June 2020

#Calculate furniture demand per quintile based on relative differences in weight between quintiles
furniture_weight = pd.DataFrame(index=furniture_weight_raw.index, columns=quintiles)

for quintile in quintiles:
    furniture_weight.loc[:, quintile] = furniture_weight_raw['Kg'].mul(weight_factor[quintile])

#furniture_weight.rename(columns={'Kitchen cabinet ':'Kitchen cabinet'}, inplace=True) # Ensuring all indexes are the same    
furniture_weight = furniture_weight.reindex(sorted(furniture_weight.columns), axis=1)
furniture_weight=furniture_weight.T
items=furniture_weight.columns #used since the files do not all contain data on the same furniture items any more
    
# Calculate furniture amount (units/m2) based on sufficiency
f_amount = pd.DataFrame(index=f_index, columns=furniture_list)

sufficiency = sufficiency.stack().reorder_levels([0,2,1]) # Reoder to the same level structure as the other data frames

for i in furniture_list:
    f_amount.loc[:,i] = sufficiency * f_amount_raw.loc[i,'Amount']

# Furniture density: kg of furniture items per square meter for each quintile (total furniture weight (kg) * units per capita per m² for each quintile)
fur_density = pd.DataFrame(index=f_amount.index, columns=items)  
fur_density = fur_density.reindex(sorted(fur_density.columns), axis=1)  
for item in items:
    for quintile in quintiles:
        fur_density.loc[idx[:,:,quintile],item] = f_amount.loc[idx[:,:,quintile],item].mul(furniture_weight.loc[quintile,item])  # kg/cap m²

fur_density.stack()         # Stack to get data in correct format to calculate total furniture demand --> doesn't do anything


# Total floorspace in millions of m2 per quintile (floorspace (m2) /cap * population (mil.)) = million m2
floorspace = pd.DataFrame(index=raw_fli_data.index, columns=quintiles)  
for quintile in quintiles:
    floorspace.loc[:, quintile] = raw_fli_data.loc[:, 'Urb_'+ quintile].multiply(pop_urb * 0.2, axis = 0) + \
                                    raw_fli_data.loc[:, 'Rur_'+ quintile].multiply(pop_rur * 0.2, axis = 0)
    # million m²
    
floorspace = floorspace.stack()   


furniture = pd.DataFrame(index=f_index, columns=items)
for quintile in quintiles:
    for item in items:
        furniture.loc[idx[:,:,quintile],item] = floorspace.loc[idx[:,:,quintile]] *fur_density.loc[idx[:,:,quintile],item]
        # stock in million kg

furniture=furniture.astype('float64')
furniture=furniture.stack()
furniture=furniture.unstack(1)
furniture=furniture.reorder_levels([2, 0, 1])
furniture=furniture.sort_index(level=0)

# splitting stock up into 1st and 2nd hand furniture
second_hand.index = furniture.loc[idx['Child bed',:,:],:].index.droplevel(0)
second_fur = pd.DataFrame().reindex_like(furniture)
first_fur = pd.DataFrame().reindex_like(furniture)

for item in items:
    for quintile in quintiles:
        second_fur.loc[idx[item,:,quintile],:] = furniture.loc[idx[item,:,quintile],:].mul(second_hand.loc[idx[:,quintile],:].values)
        first_fur.loc[idx[item,:,quintile],:]  = furniture.loc[idx[item,:,quintile],:].mul(1-second_hand.loc[idx[:,quintile],:].values)


#%% Inflow-Outflow
first_year_furniture = 1900
stdev_mult = 0.2 # to be set later

# define a dataframe that contains the weight of all furniture (stock) in kg, for every year, including the historic tail (back to 1900)
years_tail = list(range(first_year_furniture,endyear+1))
furniture_tail = pd.DataFrame(index=pd.MultiIndex.from_product([furniture_list, years_tail, quintiles], names=['t', 'Region','Quintile']), columns=regions_names)
furniture_tail = furniture_tail.astype('float64')
furniture_tail = furniture_tail.reindex(sorted(furniture_tail.columns), axis=1)

for quintile in quintiles:
    for item in items: 
        furniture_tail.loc[idx[item,:,quintile],:] = stock_tail(furniture.loc[idx[item,:,quintile],:].droplevel(level=[0,2])).to_numpy()

# define inflow and outflow according to furniture (=without historic tail) 
inflow = pd.DataFrame().reindex_like(furniture)
outflow = pd.DataFrame().reindex_like(furniture)

# then actually apply the dynamic stock model to derive inflow & outflow
for quintile in quintiles:
    for item in items:
        results = inflow_outflow(furniture_tail.loc[idx[item,:,quintile],:].droplevel(level=[0,2]), lifetimes.loc[item].values[0])
        inflow.loc[idx[item,:,quintile],:] = results[0].to_numpy()
        outflow.loc[idx[item,:,quintile],:] = results[1].to_numpy()

#%% Calculate material demand per quintile
        


#%% Figures

#%% Inflow Outflow calculations 

#%% Prepare output variables
    


#%% CSV output files